@@ -69,6 +69,54 @@ def create_normaliser(data, ranges, bins, weights = None) :
     norm += [ (cum_values, edges) ]
   return norm
 
+def create_normaliser_vector(data, ranges, bins, weights = None) : 
+  """
+  Create normaliser structure based on 2D numpy array of data. 
+
+  Normaliser is effectively a list of cumulative distributions for each dimension of 
+  the data array, that can be used to e.g. flatten or normalise the data by applying 
+  variable transformation. 
+
+  Args: 
+    data: vector of 2D numpy arrays (1st dimension is event, 2nd dimension corresponds to variable)
+    ranges: tuple of 2-element tuples corresponding to variable ranges.
+            Length of the tuple should be equal to the 2nd dimension of the data array.
+    bins: tuple of numbers of bins for each data dimension. 
+          Length of the tuple should be equal to the 2nd dimension of the data array.
+    weights: array of weights for data events (not used if None)
+
+  Returns: 
+    Normaliser structure that can be used later by reweight, normalise, unnormalise etc. 
+    functions. 
+  """
+
+  norm = []
+
+  for i in range(len(ranges)) : 
+
+    edges = None
+    counts = None
+  
+    for n,d in enumerate(data) : 
+      w = None
+      if not weights is None : w = weights[n]
+      assert d.shape[1] >= len(ranges), "shape of data array does not correspond to the number of ranges"
+      assert d.shape[1] >= len(bins), "shape of data array does not correspond to the length of the bins list"
+
+      c, e = onp.histogram(d[:,i], bins = bins[i], range = ranges[i], weights = w)
+      if counts is None : 
+        counts = c
+        edges = e
+      else : 
+        #print(counts, c)
+        counts += c
+
+    cum_values = np.concatenate( ( np.array([0.]), np.cumsum(counts*np.diff(edges)) ) )
+    cum_values /= cum_values[-1]
+    norm += [ (cum_values, edges) ]
+
+  return norm
+
 def create_histograms(data, ranges, bins, weights = None) : 
   """
   Create 1D histograms for projections of 2D numpy array of data. 
@@ -92,6 +140,42 @@ def create_histograms(data, ranges, bins, weights = None) :
     norm += [ (counts, edges) ]
   return norm
 
+def create_histograms_vector(data, ranges, bins, weights = None) : 
+  """
+  Create 1D histograms for projections of 2D numpy array of data. 
+
+  Args: 
+    data: vector of 2D numpy arrays (1st dimension is event, 2nd dimension corresponds to variable)
+    ranges: tuple of 2-element tuples corresponding to variable ranges.
+            Length of the tuple should be equal to the 2nd dimension of the data array.
+    bins: tuple of numbers of bins for each data dimension. 
+          Length of the tuple should be equal to the 2nd dimension of the data array.
+    weights: array of weights for data events (not used if None)
+
+  Returns: 
+    List of histograms, one for for each dimension of data array. 
+  """
+  norm = []
+  for i in range(len(ranges)) : 
+    edges = None
+    counts = None
+  
+    for n,d in enumerate(data) : 
+      w = None
+      if not weights is None : w = weights[n]
+      assert d.shape[1] >= len(ranges), "shape of data array does not correspond to the number of ranges"
+      assert d.shape[1] >= len(bins), "shape of data array does not correspond to the length of the bins list"
+      c, e = onp.histogram(d[:,i], bins = bins[i], range = ranges[i], weights = w)
+      if counts is None : 
+        counts = c
+        edges = e
+      else : 
+        counts += c
+
+    norm += [ (counts, edges) ]
+
+  return norm
+
 def reweight(data, norm, max_weights) : 
   """
   Calculate per-event weights for the data sample entries such that the weighted dataset is flat in 
