@@ -13,19 +13,18 @@ from ..datasets import converted_run1_pidcalib_datasets_dstar_k as ds_run1
 from ..datasets import legacy_run2_pidcalib_datasets as ds_run2
 
 from itertools import product
-from jax import numpy as np
 
 common_params = {
     "calib_cache_branches" : ["pt", "eta", "ntr", "sw"],          # Names of variable branches in ROOT tree of the cache file 
-    "calib_cache_filename" : "calib_cache.root",                  # Name of the cache ROOT file
+    "calib_cache_dirname" : "calib_cache",                  # Name of the cache ROOT file
     "data_ranges" : [ (2.4, 4.4), (2.0, 5.0), (0.7, 3.1) ],       # Ranges of each variable after transformation
     "labels" : [r"PID", r"$p_T$", r"$\eta$", r"$N_{\rm tr}$"],    # Labels for plots
     "names" : ["pid", "pt", "eta", "ntr"],                        # Short names of each variable for plot file names
     "max_weights" : None, 
     "transform" : [                                               # Transformation functions for each variable
-      lambda x : np.log10(x),   # pT transformation
-      lambda x : x,             # eta
-      lambda x : np.log10(x),   # number of tracks
+      "np.log10(x)",   # pT transformation
+      "x",             # eta
+      "np.log10(x)",   # number of tracks
     ], 
     "normalise_bins" : [100, 100, 100],                           # Number of bins for normalisation of each variable
     "normalise_methods" : ["scale", "scale", "flatten"],          # Normalisation method for each variable ("scale", "normalise", "flatten")
