@@ -12,8 +12,8 @@
 variable = {
     "branch" : "probe_Brunel_PIDp", 
     "data_range" : (-150., 150.), 
-    "transform_forward" : lambda x : x, 
-    "transform_backward" : lambda x : x, 
+    "transform_forward" : "x", 
+    "transform_backward" : "x", 
     "normalise_bins" : 1000, 
     "normalise_method" : "gauss", 
     "normalise_range" : (-2.5, 2.5), 
